@@ -11,7 +11,7 @@ import ibid
 from ibid.plugins import Processor, handler, match
 from ibid.config import Option
 from ibid.utils import file_in_path, get_country_codes, human_join, \
-                       unicode_output
+                       unicode_output, generic_webservice
 from ibid.utils.html import get_html_parse_tree
 
 features = {}
@@ -305,12 +305,15 @@ class Currency(Processor):
 
     feature = ('currency',)
 
-    headers = {'User-Agent': 'Mozilla/5.0', 'Referer': 'http://www.xe.com/'}
     currencies = {}
     country_codes = {}
 
     def _load_currencies(self):
-        etree = get_html_parse_tree('http://www.xe.com/iso4217.php', headers=self.headers, treetype='etree')
+        etree = get_html_parse_tree(
+                'http://www.xe.com/iso4217.php', headers = {
+                    'User-Agent': 'Mozilla/5.0',
+                    'Referer': 'http://www.xe.com/',
+                }, treetype='etree')
 
         tbl_main = [x for x in etree.getiterator('table') if x.get('class') == 'tbl_main'][0]
 
@@ -389,21 +392,34 @@ class Currency(Processor):
                 event.addresponse(u"Sorry, I don't know about a currency for %s", (not canonical_frm and frm or to))
             return
 
-        data = {'Amount': amount, 'From': canonical_frm, 'To': canonical_to}
-        etree = get_html_parse_tree('http://www.xe.com/ucc/convert.cgi', urlencode(data), self.headers, 'etree')
+        data = generic_webservice(
+                'http://download.finance.yahoo.com/d/quotes.csv', {
+                    'f': 'l1ba',
+                    'e': '.csv',
+                    's': canonical_frm + canonical_to + '=X',
+                })
+        last_trade_rate, bid, ask = data.strip().split(',')
+        if last_trade_rate == 0:
+            event.addresponse(
+                    u"Whoops, looks like I couldn't make that conversion")
+            return
 
-        result = [tag.text for tag in etree.getiterator('h2')]
-        if result:
-            event.addresponse(u'%(fresult)s (%(fcountry)s %(fcurrency)s) = %(tresult)s (%(tcountry)s %(tcurrency)s)', {
-                'fresult': result[0],
-                'tresult': result[2],
+        event.addresponse(
+            u'%(fresult)s %(fcode)s (%(fcountry)s %(fcurrency)s) = '
+            u'%(tresult)0.2f %(tcode)s (%(tcountry)s %(tcurrency)s) '
+            u'(Last trade rate: %(rate)s, Bid: %(bid)s, Ask: %(ask)s)', {
+                'fresult': amount,
+                'tresult': float(amount) * float(last_trade_rate),
                 'fcountry': self.currencies[canonical_frm][0][0],
                 'fcurrency': self.currencies[canonical_frm][1],
                 'tcountry': self.currencies[canonical_to][0][0],
                 'tcurrency': self.currencies[canonical_to][1],
+                'fcode': canonical_frm,
+                'tcode': canonical_to,
+                'rate': last_trade_rate,
+                'bid': bid,
+                'ask': ask,
             })
-        else:
-            event.addresponse(u"The bureau de change appears to be closed for lunch")
 
     @match(r'^(?:currency|currencies)\s+for\s+(?:the\s+)?(.+)$')
     def currency(self, event, place):
@@ -504,11 +520,11 @@ class UnicodeData(Processor):
                               u"%(category)s with %(bidi)s directionality",
                               info)
 
-    @match(r'^unicode\s+([a-z -]{2,})$')
+    @match(r'^unicode\s+([a-z][a-z0-9 -]+)$')
     def fromname (self, event, name):
         try:
             char = eval(ur'u"\N{%s}"' % name.upper())
-        except SyntaxError:
+        except (SyntaxError, UnicodeError):
             event.addresponse(u"I couldn't find a character with that name")
         else:
             info = self.info(char)
