@@ -12,6 +12,7 @@ from sys import version_info
 from urllib import getproxies_environment
 from urlparse import urlparse
 
+from chardet import detect
 try:
     from dns.resolver import Resolver, NoAnswer, NXDOMAIN
     from dns.reversename import from_address
@@ -224,12 +225,12 @@ class HTTP(Processor):
     feature = ('http',)
     priority = -10
 
-    max_size = IntOption('max_size', u'Only request this many bytes', 500)
+    max_size = IntOption('max_size', u'Only request this many bytes', 2048)
     timeout = IntOption('timeout',
             u'Timeout for HTTP connections in seconds', 15)
     sites = DictOption('sites', u'Mapping of site names to domains', {})
     max_hops = IntOption('max_hops',
-            u'Maximum hops in get/head when receiving a 30[12]', 3)
+            u'Maximum number of http redirects to follow', 5)
     whensitup_delay = IntOption('whensitup_delay',
             u'Initial delay between whensitup attempts in seconds', 60)
     whensitup_factor = FloatOption('whensitup_factor',
@@ -254,9 +255,10 @@ class HTTP(Processor):
             reply = u'%s %s' % (status, reason)
 
             hops = 0
-            while status in (301, 302) and self._get_header(headers,
-                                                            'location'):
+            while 300 <= status < 400 and self._get_header(headers, 'location'):
                 location = self._get_header(headers, 'location')
+                if not location:
+                    break
                 status, reason, data, headers = self._request(location, 'GET')
                 if hops >= self.max_hops:
                     reply += u' to %s' % location
@@ -270,10 +272,18 @@ class HTTP(Processor):
                 }
 
             if action.upper() == 'GET':
-                match = re.search(r'<title>(.*)<\/title>', data,
-                                  re.I | re.DOTALL)
-                if match:
-                    reply += u' "%s"' % match.groups()[0].strip()
+                got_title = False
+                content_type = self._get_header(headers, 'content-type')
+                if content_type and (content_type.startswith('text/html') or
+                        content_type.startswith('application/xhtml+xml')):
+                    match = re.search(r'<title>(.*)<\/title>', data,
+                                      re.I | re.DOTALL)
+                    if match:
+                        got_title = True
+                        reply += u' "%s"' % match.groups()[0].strip()
+
+                if not got_title and content_type:
+                    reply += u' ' + content_type
 
             event.addresponse(reply)
 
@@ -285,14 +295,15 @@ class HTTP(Processor):
             url = self.sites[url]
         else:
             if not urlparse(url).netloc:
-                if '.' not in url:
+                if '.' not in url and '/' not in url and ':' not in url:
+                    # Only append .com for url's like "example"
                     url += '.com'
                 url = 'http://' + url
             if not urlparse(url).path:
                 url += '/'
         return url
 
-    def _isitup(self, url):
+    def _isitup(self, url, return_status=False, redirects=0):
         valid_url = self._makeurl(url)
         if Resolver is not None:
             r = Resolver()
@@ -300,29 +311,33 @@ class HTTP(Processor):
             try:
                 r.query(host)
             except NoAnswer:
-                return False, u'No DNS A/CNAME-records for that domain'
+                return (False, valid_url,
+                        u'No DNS A/CNAME-records for that domain')
             except NXDOMAIN:
-                return False, u'No such domain'
+                return False, valid_url, u'No such domain'
 
         try:
             status, reason, data, headers = self._request(valid_url, 'HEAD')
-            # If the URL is only a hostname, we consider 400 series errors to
-            # mean up. Full URLs are checked for a sensible response code
-            if url == urlparse(url).path and '/' not in url:
-                up = status < 500
-            else:
-                up = status < 400
+            if 300 <= status < 400 and self._get_header(headers, 'location'):
+                if redirects > self.max_hops:
+                    return False, valid_url, u'Redirect limit reached'
+                return self._isitup(self._get_header(headers, 'location'),
+                                    return_status, redirects + 1)
+
+            up = status < 300
+            if return_status:
                 reason = u'%(status)d %(reason)s' % {
                     u'status': status,
                     u'reason': reason,
                 }
-            return up, reason
+            return up, valid_url, reason
         except HTTPException:
-            return False, u'Server is not responding'
+            return False, valid_url, u'Server is not responding'
 
     @match(r'^is\s+(\S+)\s+(up|down)$')
     def isit(self, event, url, type):
-        up, reason = self._isitup(url)
+        host_only = (url == urlparse(url).path and '/' not in url)
+        up, url, reason = self._isitup(url, return_status=not host_only)
         if up:
             if type.lower() == 'up':
                 event.addresponse(u'Yes, %s is up', url)
@@ -350,6 +365,7 @@ class HTTP(Processor):
             event.addresponse(u"Sorry, it appears %s is never coming up. "
                               u"I'm not going to check any more.",
                               self._makeurl(url))
+            return
         delay *= self.whensitup_factor
         delay = max(delay, self.whensitup_maxdelay)
         ibid.dispatcher.call_later(delay, self._whensitup, event, url, delay,
@@ -402,13 +418,24 @@ class HTTP(Processor):
             if version_info[1] < 6:
                 socket.setdefaulttimeout(None)
 
-        contenttype = response.getheader('Content-Type',
-                                         'text/html; charset=utf-8')
-        match = re.search('charset=([a-zA-Z0-9-]+)', contenttype)
-        charset = match and match.group(1) or 'utf-8'
+        contenttype = response.getheader('Content-Type', None)
+        if contenttype:
+            match = re.search('^charset=([a-zA-Z0-9-]+)', contenttype)
+            try:
+                if match:
+                    data = data.decode(match.group(1))
+                elif contenttype.startswith('text/'):
+                    data = data.decode('utf-8')
+            except UnicodeDecodeError:
+                guessed = detect(data)
+                if guessed['confidence'] > 0.5:
+                    charset = guessed['encoding']
+                    # Common guessing mistake:
+                    if charset.startswith('ISO-8859') and '\x92' in data:
+                        charset = 'windows-1252'
+                    data = unicode(data, charset, errors='replace')
 
-        return response.status, response.reason, data.decode(charset), \
-               response.getheaders()
+        return response.status, response.reason, data, response.getheaders()
 
 features['tld'] = {
     'description': u'Resolve country TLDs (ISO 3166)',
